import {IBackendConnector} from "./IBackendConnector.ts";
import {CustomPlantPage, CustomPlantPageSchema, Plant, PlantSchema, SearchRequest} from "./schemas.ts";
import {ZodError} from "zod";
import {FetchResult} from "./FetchResult.ts";
import {HttpMethod} from "./commonTypes";
import {getCookie} from "../jwt/Cookies.ts";

export class BackendConnectorImpl implements IBackendConnector {
    private readonly baseUrl: string;

    constructor(baseUrl: string) {
        const lastSlash = baseUrl.lastIndexOf("/");
        const endsWithSlash = lastSlash === baseUrl.length - 1;
        if (endsWithSlash) {
            baseUrl = baseUrl.substring(0, lastSlash)
        }
        this.baseUrl = baseUrl;
    }

    async getAllPlants(page?: number, pageSize?: number): Promise<FetchResult<CustomPlantPage>> {
        let result: FetchResult<CustomPlantPage>;
        try {
            const pageParam = page ? `page=${page}&` : "";
            const pageSizeParam = pageSize ? `pageSize=${pageSize}` : "";
            const response = await this.makeFetchWithAuth("/plants?" + pageParam + pageSizeParam);
            const json = await response.json();
            result = this.parsePlantsFromPage(json);
        } catch (e) {
            console.error("An error occurred while fetching all plants")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async searchPlants(search: SearchRequest): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]>;
        try {
            const searchParams = new URLSearchParams(
                Object.entries(search)
                    .map(([key, value]) => [key, value as string]));
            const response = await this.makeFetchWithAuth("/plants/search?" + searchParams);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while searching through plants")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getRandomPlants(amount: number): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]>;
        try {
            const response = await this.makeFetchWithAuth("/plants/randomPlants/" + amount);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while getting random plants")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getSinglePlant(id: number): Promise<FetchResult<Plant | null>> {
        let result: FetchResult<Plant>;
        try {
            const response = await this.makeFetchWithAuth("/plants/" + id);
            const json = await response.json();
            result = this.parseSinglePlant(json);
        } catch (e) {
            console.error("An error occurred while fetching plant with ID", id);
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async addToGarden(plantId: number): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {value: undefined};
        try {
            await this.makeFetchWithAuthAndOptions("/garden/add/" + plantId, {method: "POST"});
        } catch (e) {
            console.error("An error occurred while adding plant with ID", plantId, "to garden");
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async clearGarden(): Promise<FetchResult<number>> {
        let result: FetchResult<number>;
        try {
            const res = await this.makeFetchWithAuthAndOptions("/garden/remove/all", {method: "DELETE"});
            const num = await res.json();
            result = {value: num};
        } catch (e) {
            console.error("An error occurred while clearing garden");
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getGardenEntries(sort?: string): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]>;
        try {
            const params = new URLSearchParams();
            if (sort) params.set("sort", sort);
            const response = await this.makeFetchWithAuth("/garden?" + params);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while getting garden entries")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getGardenSize(): Promise<FetchResult<number>> {
        let result: FetchResult<number>;
        try {
            const response = await this.makeFetchWithAuth("/garden/count");
            const json = await response.json();
            const size = Number.parseInt(json);
            result = {value: size};
        } catch (e) {
            console.error("An error occurred while fetching garden size")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async removeEntryFromGarden(entryId: number): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {value: undefined};
        try {
            await this.makeFetchWithAuthAndOptions("/garden/remove/" + entryId, {method: "DELETE"});
        } catch (e) {
            console.error("An error occurred while removing garden entry with ID", entryId);
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async addToWishlist(plantId: number): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {value: undefined};
        try {
            await this.makeFetchWithAuthAndOptions("/wishlist/add/" + plantId, {method: "POST"});
        } catch (e) {
            console.error("An error occurred while adding plant to wishlist with ID", plantId);
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async clearWishlist(): Promise<FetchResult<number>> {
        let result: FetchResult<number>;
        try {
            const res = await this.makeFetchWithAuthAndOptions("/wishlist/remove/all", {method: "DELETE"});
            const num = await res.json();
            result = {value: num}
        } catch (e) {
            console.error("An error occurred while clearing wishlist");
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getWishlistEntries(sort?: string): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]>;
        try {
            const params = new URLSearchParams();
            if (sort) params.set("sort", sort);
            const response = await this.makeFetchWithAuth("/wishlist?" + params);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while getting wishlist entries")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getWishlistSize(): Promise<FetchResult<number>> {
        let result: FetchResult<number>;
        try {
            const response = await this.makeFetchWithAuth("/wishlist/count");
            const json = await response.json();
            const size = Number.parseInt(json);
            result = {value: size};
        } catch (e) {
            console.error("An error occurred while fetching wishlist size")
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async removeFromWishlist(plantId: number): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {value: undefined};
        try {
            await this.makeFetchWithAuthAndOptions("/wishlist/remove/" + plantId, {method: "DELETE"});
        } catch (e) {
            console.error("An error occurred while removing entry from wishlist with id", plantId)
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    private makeFetchWithAuth(path: string): Promise<Response> {
        return this.makeFetchWithAuthAndOptions(path, {});
    }

    private makeFetchWithAuthAndOptions(path: string, options: {
        method?: HttpMethod,
        body?: object
    }): Promise<Response> {
        const jwt = getCookie("Jwt");
        if (!jwt) {
            throw new Error("no jwt found for call to backend");
        }
        console.log("Making request to", path)
        const {method, body} = options;
        let headers: HeadersInit = {
            Authorization: `Bearer ${jwt}`
        };
        if (body) {
            headers = {...headers, contentType: "application/json"}
        }
        return fetch(this.baseUrl + path, {
            method: method,
            headers: headers,
            body: JSON.stringify(body),
        });
    }

    private parseSinglePlant(json: any): FetchResult<Plant> {
        let result: FetchResult<Plant>;
        try {
            result = {value: PlantSchema.parse(json)};
        } catch (e) {
            result = {err: "One or more plants could not be parsed correctly."}
        }
        return result;
    }

    private parsePlants(json: any): FetchResult<Plant[]> {
        let result: FetchResult<Plant[]> = {value: []};
        const plants: Plant[] = [];
        for (const p of json) {
            try {
                plants.push(PlantSchema.parse(p))
            } catch (e) {
                if (e instanceof ZodError) {
                    result = {...result, err: "One or more plants could not be parsed correctly."}
                }
            }
        }
        return {...result, value: plants};
    }

    private parsePlantsFromPage(json: any): FetchResult<CustomPlantPage> {
        let result: FetchResult<CustomPlantPage>;
        try {
            result = {value: CustomPlantPageSchema.parse(json)}
        } catch (e) {
            result = {err: "One or more plants could not be parsed correctly."}
        }
        return result;
    }
}
